from py.test import raises

from graphene import Field, Int, Interface, ObjectType
from graphene.relay import Node, is_node

from .. import registry
from ..types import MongoengineObjectType
from .models import Article
from .models import Reporter
from .utils import with_local_registry

registry.reset_global_registry()


class Human(MongoengineObjectType):

    pub_date = Int()

    class Meta:
        model = Article
        registry = registry.get_global_registry()
        interfaces = (Node,)


class Character(MongoengineObjectType):
    class Meta:
        model = Reporter
        registry = registry.get_global_registry()


def test_mongoengine_interface():
    assert issubclass(Node, Interface)
    assert issubclass(Node, Node)


def test_objecttype_registered():
    assert issubclass(Character, ObjectType)
    assert Character._meta.model == Reporter
    assert set(Character._meta.fields.keys()) == set([
        'id',
        'first_name',
        'last_name',
        'email',
        # FIXME
        # 'embedded_articles',
        'articles',
        'awards'
    ])


def test_node_replacedfield():
    idfield = Human._meta.fields['pub_date']
    assert isinstance(idfield, Field)
    assert idfield.type == Int


def test_object_type():
    assert issubclass(Human, ObjectType)
    assert set(Human._meta.fields.keys()) == set([
        'id',
        'headline',
        'pub_date',
        'editor',
        'reporter'
    ])
    assert is_node(Human)


def test_should_raise_if_no_model():
    with raises(Exception) as excinfo:
        class Human1(MongoengineObjectType):
            pass
    assert 'valid Mongoengine Model' in str(excinfo.value)


def test_should_raise_if_model_is_invalid():
    with raises(Exception) as excinfo:
        class Human2(MongoengineObjectType):
            class Meta:
                model = 1
    assert 'valid Mongoengine Model' in str(excinfo.value)


@with_local_registry
def test_mongoengine_objecttype_only_fields():
    class A(MongoengineObjectType):
        class Meta:
            model = Article
            only_fields = ('headline')


    fields = set(A._meta.fields.keys())
    assert fields == set(['headline'])


@with_local_registry
def test_mongoengine_objecttype_exclude_fields():
    class A(MongoengineObjectType):
        class Meta:
            model = Article
            exclude_fields = ('headline')

    assert 'headline' not in list(A._meta.fields.keys())




import inspect
import mongoengine

from collections import OrderedDict
from mongoengine.base.fields import BaseField


def get_model_fields(model, excluding=None):
    if excluding is None:
        excluding = []
    attributes = dict()
    for attr_name in vars(model):
        if attr_name in excluding:
            continue
        attr = getattr(model, attr_name)
        if isinstance(attr, BaseField):
            attributes[attr_name] = attr

    return OrderedDict(sorted(attributes.items()))


def is_valid_mongoengine_model(model):
    return inspect.isclass(model) and (
        issubclass(model, mongoengine.Document) or issubclass(model, mongoengine.EmbeddedDocument)
    )


def import_single_dispatch():
    try:
        from functools import singledispatch
    except ImportError:
        singledispatch = None

    if not singledispatch:
        try:
            from singledispatch import singledispatch
        except ImportError:
            pass

    if not singledispatch:
        raise Exception(
            "It seems your python version does not include "
            "functools.singledispatch. Please install the 'singledispatch' "
            "package. More information here: "
            "https://pypi.python.org/pypi/singledispatch"
        )

    return singledispatch


# noqa
def get_type_for_document(schema, document):
    types = schema.types.values()
    for _type in types:
        type_document = hasattr(_type, '_meta') and getattr(
            _type._meta, 'document', None)
        if document == type_document:
            return _type



from graphene import (ID, Boolean, Dynamic, Field, Float, Int, List,
                      NonNull, String, is_node)
from graphene.types.json import JSONString

import mongoengine

from .fields import MongoengineConnectionField
from .utils import import_single_dispatch

singledispatch = import_single_dispatch()


@singledispatch
def convert_mongoengine_field(field, registry=None):
    raise Exception(
        "Don't know how to convert the MongoEngine field %s (%s)" %
        (field, field.__class__))


@convert_mongoengine_field.register(mongoengine.EmailField)
@convert_mongoengine_field.register(mongoengine.StringField)
@convert_mongoengine_field.register(mongoengine.URLField)
def convert_field_to_string(field, registry=None):
    return String(description=field.db_field, required=not field.null)


@convert_mongoengine_field.register(mongoengine.UUIDField)
@convert_mongoengine_field.register(mongoengine.ObjectIdField)
def convert_field_to_id(field, registry=None):
    return ID(description=field.db_field, required=not field.null)


@convert_mongoengine_field.register(mongoengine.IntField)
def convert_field_to_int(field, registry=None):
    return Int(description=field.db_field, required=not field.null)


@convert_mongoengine_field.register(mongoengine.BooleanField)
def convert_field_to_boolean(field, registry=None):
    return NonNull(Boolean, description=field.db_field)


@convert_mongoengine_field.register(mongoengine.DecimalField)
@convert_mongoengine_field.register(mongoengine.FloatField)
def convert_field_to_float(field, registry=None):
    return Float(description=field.db_field, required=not field.null)


@convert_mongoengine_field.register(mongoengine.DictField)
@convert_mongoengine_field.register(mongoengine.MapField)
def convert_dict_to_jsonstring(field, registry=None):
    return JSONString(description=field.db_field, required=not field.null)


@convert_mongoengine_field.register(mongoengine.DateTimeField)
def convert_date_to_string(field, registry=None):
    return String(description=field.db_field, required=not field.null)


@convert_mongoengine_field.register(mongoengine.ListField)
@convert_mongoengine_field.register(mongoengine.EmbeddedDocumentListField)
def convert_field_to_list(field, registry=None):
    base_type = convert_mongoengine_field(field.field, registry=registry)
    if isinstance(base_type, (Dynamic)):
        base_type = base_type.get_type()
        if base_type is None:
            return
        base_type = base_type._type

    if is_node(base_type):
        return MongoengineConnectionField(base_type)

    # Non-relationship field
    relations = (mongoengine.ReferenceField, mongoengine.EmbeddedDocumentField)
    if not isinstance(base_type, (List, NonNull)) \
            and not isinstance(field.field, relations):
        base_type = type(base_type)

    return List(base_type, description=field.db_field, required=not field.null)


@convert_mongoengine_field.register(mongoengine.EmbeddedDocumentField)
@convert_mongoengine_field.register(mongoengine.ReferenceField)
def convert_field_to_dynamic(field, registry=None):
    model = field.document_type

    def dynamic_type():
        _type = registry.get_type_for_model(model)
        if not _type:
            return None

        return Field(_type)

    return Dynamic(dynamic_type)



import pytest
from mongoengine import Document


from ..utils import (
    get_model_fields, is_valid_mongoengine_model
)
from .models import Reporter

def test_get_model_fields_no_duplication():
    reporter_fields = get_model_fields(Reporter)
    reporter_name_set = set(reporter_fields)
    assert len(reporter_fields) == len(reporter_name_set)


def test_is_valid_mongoengine_mode():
    assert is_valid_mongoengine_model(Reporter)




from __future__ import absolute_import

from collections import OrderedDict
from functools import partial, reduce

from graphene import Field, List
from graphene.relay import ConnectionField
from graphene.relay.connection import PageInfo
from graphql_relay.connection.arrayconnection import connection_from_list_slice
from graphql_relay.node.node import from_global_id
from graphene.types.argument import to_arguments


# noqa
class MongoengineListField(Field):

    def __init__(self, _type, *args, **kwargs):
        super(MongoengineListField, self).__init__(
            List(_type), *args, **kwargs)

    @property
    def model(self):
        return self.type.of_type._meta.node._meta.model

    # @staticmethod
    # def list_resolver(resolver, root, info, **args):
    #    return maybe_queryset(resolver(root, info, **args))

    def get_resolver(self, parent_resolver):
        return partial(self.list_resolver, parent_resolver)


class MongoengineConnectionField(ConnectionField):

    def __init__(self, type, *args, **kwargs):
        super(MongoengineConnectionField, self).__init__(
            type,
            *args,
            **kwargs
        )

    @property
    def type(self):
        from .types import MongoengineObjectType
        _type = super(ConnectionField, self).type
        assert issubclass(
            _type, MongoengineObjectType), "MongoengineConnectionField only accepts MongoengineObjectType types"
        assert _type._meta.connection, "The type {} doesn't have a connection".format(
            _type.__name__)
        return _type._meta.connection

    @property
    def node_type(self):
        return self.type._meta.node

    @property
    def model(self):
        return self.node_type._meta.model

    @property
    def args(self):
        return to_arguments(
            self._base_args or OrderedDict(), self.default_filter_args
        )

    @args.setter
    def args(self, args):
        self._base_args = args

    @property
    def default_filter_args(self):
        def is_filterable(kv):
            return hasattr(kv[1], '_type') \
                and callable(getattr(kv[1]._type, '_of_type', None))

        return reduce(
            lambda r, kv: r.update(
                {kv[0]: kv[1]._type._of_type()}) or r if is_filterable(kv) else r,
            self.fields.items(),
            {}
        )

    @property
    def filter_fields(self):
        return self._type._meta.filter_fields

    @property
    def fields(self):
        return self._type._meta.fields

    @classmethod
    def get_query(cls, model, info, **args):

        if not callable(getattr(model, 'objects', None)):
            return []

        objs = model.objects()

        if args:
            first = args.pop('first', None)
            last = args.pop('last', None)
            id = args.pop('id', None)
            before = args.pop('before', None)
            after = args.pop('after', None)

            if id is not None:
                # https://github.com/graphql-python/graphene/issues/124
                args['pk'] = from_global_id(id)[-1]

            objs = objs.filter(**args)

            # https://github.com/graphql-python/graphene-mongo/issues/21
            if after is not None:
                _after = int(from_global_id(after)[-1])
                objs = objs[_after:]

            if before is not None:
                _before = int(from_global_id(before)[-1])
                objs = objs[:_before]

            if first is not None:
                objs = objs[:first]
            if last is not None:
                # fix for https://github.com/graphql-python/graphene-mongo/issues/20
                objs = objs[-(last+1):]

        return objs

    # noqa
    @classmethod
    def merge_querysets(cls, default_queryset, queryset):
        return queryset & default_queryset

    """
    Notes: Not sure how does this work :(
    """
    @classmethod
    def connection_resolver(cls, resolver, connection, model, root, info, **args):
        iterable = resolver(root, info, **args)
        if not iterable:
            iterable = cls.get_query(model, info, **args)
        _len = len(iterable)
        connection = connection_from_list_slice(
            iterable,
            args,
            slice_start=0,
            list_length=_len,
            list_slice_length=_len,
            connection_type=connection,
            pageinfo_type=PageInfo,
            edge_type=connection.Edge,
        )
        connection.iterable = iterable
        connection.length = _len
        return connection

    def get_resolver(self, parent_resolver):
        return partial(self.connection_resolver, parent_resolver, self.type, self.model)



from .models import Article, Editor, EmbeddedArticle, Player, Reporter


def setup_fixtures():
    Editor.drop_collection()
    editor1 = Editor(
        id='1',
        first_name='Penny',
        last_name='Hardaway',
        metadata={'age': '20', 'nickname': '$1'}
    )
    editor1.save()
    editor2 = Editor(
        id='2',
        first_name='Grant',
        last_name='Hill'
    )
    editor2.save()
    editor3 = Editor(
        id='3',
        first_name='Dennis',
        last_name='Rodman'
    )
    editor3.save()

    Article.drop_collection()
    article1 = Article(headline='Hello', editor=editor1)
    article1.save()
    article2 = Article(headline='World', editor=editor2)
    article2.save()

    Reporter.drop_collection()
    reporter1 = Reporter(
        id='1',
        first_name='Allen',
        last_name='Iverson',
        email='ai@gmail.com',
        awards=['2010-mvp']
    )
    reporter1.articles = [article1, article2]
    embedded_article1 = EmbeddedArticle(
        headline='Real',
        editor=editor1
    )
    embedded_article2 = EmbeddedArticle(
        headline='World',
        editor=editor2
    )
    reporter1.embedded_articles = [embedded_article1, embedded_article2]
    reporter1.embedded_list_articles = [embedded_article2, embedded_article1]
    reporter1.save()

    Player.drop_collection()
    player1 = Player(
        first_name='Michael',
        last_name='Jordan'
    )
    player1.save()
    player2 = Player(
        first_name='Magic',
        last_name='Johnson',
        opponent=player1
    )
    player2.save()
    player3 = Player(
        first_name='Larry',
        last_name='Bird',
        players=[player1, player2]
    )
    player3.save()

    player1.players = [player2]
    player1.save()

    player2.players = [player1]
    player2.save()
